"""Test the pytest plugin."""
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=unused-argument
import pytest


@pytest.fixture
def tmp_conftest(ref_tree, res_tree_equal, ref_csv, res_csv_equal):
    """Create a temporary conftest file."""
    return f"""
        from pathlib import Path

        import pytest

        @pytest.fixture
        def ref_path():
            return Path("{ref_tree}")

        @pytest.fixture
        def res_path():
            return Path("{res_tree_equal}")
        """


@pytest.mark.parametrize(
    "do_export, export_suffix",
    [
        [False, None],
        [False, "_CMD_SUFFIX"],
        [True, None],
        [True, "_CMD_SUFFIX"],
    ],
)
def test_export_formatted_data(
    ref_tree,
    res_tree_equal,
    ref_csv,
    res_csv_equal,
    tmp_conftest,
    pytester,
    do_export,
    export_suffix,
    registry_reseter,
):
    """Test that the formatted files are properly exported."""
    args = []
    if not do_export:
        expected_dir = """res_path.with_name(res_path.name + "_FORMATTED")"""
        tester = f"""assert not {expected_dir}.exists()"""
    else:
        if export_suffix is None:
            suffix = "_FORMATTED"
        else:
            suffix = export_suffix

        expected_dir = f"""res_path.with_name(res_path.name + "{suffix}")"""
        args.append("--dcd-export-formatted-data")
        if export_suffix is not None:
            args.append("--dcd-export-suffix")
            args.append(export_suffix)
        tester = """assert list(expected_dir.iterdir()) == [expected_dir / "file.csv"]"""

    expected_dir_str = f"""expected_dir = {expected_dir}"""
    remover = """rmtree(expected_dir, ignore_errors=True)"""

    # create a temporary conftest.py file
    pytester.makeconftest(tmp_conftest)

    # create a temporary pytest test file
    pytester.makepyfile(
        f"""
        from shutil import rmtree

        import dir_content_diff
        import dir_content_diff.pandas
        from dir_content_diff import assert_equal_trees

        dir_content_diff.reset_comparators()
        dir_content_diff.pandas.register()


        def test_export_formatted_data_default(ref_path, res_path):
            {expected_dir_str}
            {remover}
            assert_equal_trees(ref_path, res_path)
            {tester}


        def test_export_formatted_data_no_suffix(ref_path, res_path):
            expected_dir = res_path.with_name(res_path.name + "_FORMATTED")
            rmtree(expected_dir, ignore_errors=True)

            assert_equal_trees(ref_path, res_path, export_formatted_files=True)
            assert list(expected_dir.iterdir()) == [expected_dir / "file.csv"]


        def test_export_formatted_data_suffix(ref_path, res_path):
            expected_dir = res_path.with_name(res_path.name + "_NEW_SUFFIX")
            rmtree(expected_dir, ignore_errors=True)

            assert_equal_trees(ref_path, res_path, export_formatted_files="_NEW_SUFFIX")
            assert list(expected_dir.iterdir()) == [expected_dir / "file.csv"]

        """
    )

    # run all tests with pytest
    result = pytester.runpytest(*args)

    # check that all 3 tests passed
    result.assert_outcomes(passed=3)



"""dir-content-diff package.

Simple tool to compare directory contents.
"""
import copy
import importlib.metadata
import logging
from pathlib import Path

from dir_content_diff.base_comparators import DefaultComparator
from dir_content_diff.base_comparators import JsonComparator
from dir_content_diff.base_comparators import PdfComparator
from dir_content_diff.base_comparators import XmlComparator
from dir_content_diff.base_comparators import YamlComparator
from dir_content_diff.util import diff_msg_formatter
from dir_content_diff.util import format_ext

__version__ = importlib.metadata.version("dir-content-diff")

L = logging.getLogger(__name__)


_DEFAULT_COMPARATORS = {
    None: DefaultComparator(),
    ".json": JsonComparator(),
    ".pdf": PdfComparator(),
    ".yaml": YamlComparator(),
    ".yml": YamlComparator(),
    ".xml": XmlComparator(),
}

_DEFAULT_EXPORT_SUFFIX = "_FORMATTED"

_COMPARATORS = {}


def reset_comparators():
    """Reset the comparator registry to the default values."""
    global _COMPARATORS  # pylint: disable=global-statement
    _COMPARATORS = copy.deepcopy(_DEFAULT_COMPARATORS)


reset_comparators()


def get_comparators():
    """Return a copy of the comparator registry."""
    return copy.deepcopy(_COMPARATORS)


def register_comparator(ext, comparator, force=False):
    """Add a comparator to the registry.

    Args:
        ext (str): The extension to register.
        comparator (callable): The comparator that should be associated with the given extension.
        force (bool): If set to ``True``, no exception is raised if the given ``ext`` is already
            registered and the comparator is replaced.

    .. note::
        The given comparator should have the following signature:

        .. code-block:: python

            comparator(
                ref_file: str,
                comp_file: str,
                *diff_args: Sequence[Any],
                return_raw_diffs: bool=False,
                **diff_kwargs: Mapping[str, Any],
            ) -> Union[False, str]

        The return type can be Any when used with `return_raw_diffs == True`.
    """
    ext = format_ext(ext)
    if not force and ext in _COMPARATORS:
        raise ValueError(
            f"The '{ext}' extension is already registered and must be unregistered before being "
            "replaced."
        )
    _COMPARATORS[ext] = comparator


def unregister_comparator(ext, quiet=False):
    """Remove a comparator from the registry.

    Args:
        ext (str): The extension to unregister.
        quiet (bool): If set to ``True``, no exception is raised if the given ``ext`` is not
            registered.

    Returns:
        The removed comparator.
    """
    ext = format_ext(ext)
    if not quiet and ext not in _COMPARATORS:
        raise ValueError(f"The '{ext}' extension is not registered.")
    return _COMPARATORS.pop(ext, None)


def compare_files(ref_file, comp_file, comparator, *args, return_raw_diffs=False, **kwargs):
    """Compare 2 files and return the difference.

    Args:
        ref_file (str): Path to the reference file.
        comp_file (str): Path to the compared file.
        comparator (callable): The comparator to use (see in :func:`register_comparator` for the
            comparator signature).
        return_raw_diffs (bool): If set to ``True``, only the raw differences are returned instead
            of a formatted report.
        *args: passed to the comparator.
        **kwargs: passed to the comparator.

    Returns:
        bool or str: ``False`` if the files are equal or a string with a message explaining the
        differences if they are different.
    """
    # Get the compared file
    L.debug("Compare: %s and %s", ref_file, comp_file)

    try:
        return comparator(ref_file, comp_file, *args, return_raw_diffs=return_raw_diffs, **kwargs)
    except Exception as exception:  # pylint: disable=broad-except
        load_kwargs = kwargs.pop("load_kwargs", None)
        format_data_kwargs = kwargs.pop("format_data_kwargs", None)
        filter_kwargs = kwargs.pop("filter_kwargs", None)
        format_diff_kwargs = kwargs.pop("format_diff_kwargs", None)
        sort_kwargs = kwargs.pop("sort_kwargs", None)
        concat_kwargs = kwargs.pop("concat_kwargs", None)
        report_kwargs = kwargs.pop("report_kwargs", None)
        return diff_msg_formatter(
            ref_file,
            comp_file,
            reason="Exception raised: " + "\n".join(exception.args),
            diff_args=args,
            diff_kwargs=kwargs,
            load_kwargs=load_kwargs,
            format_data_kwargs=format_data_kwargs,
            filter_kwargs=filter_kwargs,
            format_diff_kwargs=format_diff_kwargs,
            sort_kwargs=sort_kwargs,
            concat_kwargs=concat_kwargs,
            report_kwargs=report_kwargs,
        )


def export_formatted_file(file, formatted_file, comparator, **kwargs):
    """Format a data file and export it.

    .. note:: A new file is created only if the corresponding comparator has saving capability.

    Args:
        file (str): Path to the compared file.
        formatted_file (str): Path to the formatted file.
        comparator (callable): The comparator to use (see in :func:`register_comparator` for the
            comparator signature).
        **kwargs: Can contain the following dictionaries: 'load_kwargs', 'format_data_kwargs' and
            'save_kwargs'.
    """
    if hasattr(comparator, "save_capability") and comparator.save_capability:
        # pylint: disable=protected-access
        L.debug("Format: %s into %s", file, formatted_file)
        data = comparator.load(
            file,
            **kwargs.get(
                "load_kwargs",
                comparator._default_load_kwargs
                if hasattr(comparator, "_default_load_kwargs")
                else {},
            ),
        )
        formatted_data = comparator.format_data(
            data,
            **kwargs.get(
                "format_data_kwargs",
                comparator._default_format_data_kwargs
                if hasattr(comparator, "_default_format_data_kwargs")
                else {},
            ),
        )
        Path(formatted_file).parent.mkdir(parents=True, exist_ok=True)
        comparator.save(
            formatted_data,
            formatted_file,
            **kwargs.get(
                "save_kwargs",
                comparator._default_save_kwargs
                if hasattr(comparator, "_default_save_kwargs")
                else {},
            ),
        )
    else:
        L.info("Skip formatting for '%s' because the comparator has no saving capability.", file)


def compare_trees(
    ref_path,
    comp_path,
    comparators=None,
    specific_args=None,
    return_raw_diffs=False,
    export_formatted_files=False,
):
    """Compare all files from 2 different directory trees and return the differences.

    .. note::

        The comparison only considers the files found in the reference directory. So if there are
        files in the compared directory that do not exist in the reference directory, they are just
        ignored.

    Args:
        ref_path (str): Path to the reference directory.
        comp_path (str): Path to the directory that must be compared against the reference.
        comparators (dict): A ``dict`` to override the registered comparators.
        specific_args (dict): A ``dict`` with the args/kwargs that should be given to the
            comparator for a given file. This ``dict`` should be like the following:

            .. code-block:: Python

                {
                    <relative_file_path>: {
                        args: [arg1, arg2, ...],
                        kwargs: {
                            kwarg_name_1: kwarg_value_1,
                            kwarg_name_2: kwarg_value_2,
                        }
                    },
                    <another_file_path>: {...}
                }
        return_raw_diffs (bool): If set to ``True``, only the raw differences are returned instead
            of a formatted report.
        export_formatted_files (bool or str): If set to ``True`` or a not empty string, create a
            new directory with formatted compared data files. If a string is passed, this string is
            used as suffix for the new directory. If `True` is passed, the suffix is
            ``_FORMATTED``.

    Returns:
        dict: A ``dict`` in which the keys are the relative file paths and the values are the
        difference messages. If the directories are considered as equal, an empty ``dict`` is
        returned.
    """
    if comparators is None:
        comparators = _COMPARATORS

    ref_path = Path(ref_path)
    comp_path = Path(comp_path)
    formatted_data_path = comp_path.with_name(
        comp_path.name
        + (
            export_formatted_files
            if (export_formatted_files is not True and export_formatted_files)
            else _DEFAULT_EXPORT_SUFFIX
        )
    )

    if specific_args is None:
        specific_args = {}

    # Loop over all files and call the correct comparator
    different_files = {}
    for ref_file in ref_path.glob("**/*"):
        if ref_file.is_dir():
            continue

        relative_path = ref_file.relative_to(ref_path).as_posix()
        comp_file = comp_path / relative_path

        if comp_file.exists():
            specific_file_args = specific_args.get(relative_path, {})
            comparator = comparators.get(ref_file.suffix, _COMPARATORS.get(None))
            comparator_kwargs = {k: v for k, v in specific_file_args.items() if k != "args"}
            res = compare_files(
                ref_file,
                comp_file,
                comparator,
                *specific_file_args.get("args", []),
                return_raw_diffs=return_raw_diffs,
                **comparator_kwargs,
            )
            if res is not False:
                different_files[relative_path] = res
            if export_formatted_files is not False:
                export_formatted_file(
                    comp_file,
                    formatted_data_path / relative_path,
                    comparator,
                    **comparator_kwargs,
                )
        else:
            msg = f"The file '{relative_path}' does not exist in '{comp_path}'."
            different_files[relative_path] = msg

    return different_files


def assert_equal_trees(*args, export_formatted_files=False, **kwargs):
    """Raise an :class:`AssertionError` if differences are found in the two directory trees.

    .. note::
        This function has a specific behavior when run with pytest. See the doc of the
        :mod:`dir_content_diff.pytest_plugin`.

    Args:
        *args: passed to the :func:`compare_trees` function.
        export_formatted_files (bool, or str): If set to ``True``, the formatted files are exported
            to the directory with the default suffix. If set to a string, it is used as suffix for
            the new directory.
        **kwargs: passed to the :func:`compare_trees` function.

    Returns:
        (bool) ``True`` if the trees are equal. If they are not, an :class:`AssertionError` is
        raised.
    """
    # If run with pytest, get the trigger to export formatted data from it
    if export_formatted_files is False and hasattr(
        assert_equal_trees, "_pytest_export_formatted_data"
    ):
        # pylint: disable=no-member
        # pylint: disable=protected-access
        export_formatted_files = assert_equal_trees._pytest_export_formatted_data
        if export_formatted_files is True and assert_equal_trees._pytest_export_suffix:
            export_formatted_files = assert_equal_trees._pytest_export_suffix

    different_files = compare_trees(*args, export_formatted_files=export_formatted_files, **kwargs)

    # Sort the files according to their relative paths
    sorted_items = sorted(different_files.items(), key=lambda x: x[0])
    # Test that all files are equal and raise the formatted messages if there are differences
    if len(sorted_items) > 0:
        raise AssertionError("\n\n\n".join([i[1] for i in sorted_items]))

    # Return True if the trees are equal
    return True



"""Configuration for the pytest test suite."""
# pylint: disable=redefined-outer-name
from pathlib import Path

import pandas as pd
import pytest

import dir_content_diff

from . import generate_test_files

pytest_plugins = ["pytester"]


@pytest.fixture
def registry_reseter():
    """Fixture to automatically reset the registry before and after a test."""
    dir_content_diff.reset_comparators()
    yield None
    dir_content_diff.reset_comparators()


@pytest.fixture
def empty_ref_tree(tmpdir):
    """Empty reference directory."""
    tree = Path(tmpdir) / "ref"
    tree.mkdir()
    return tree


@pytest.fixture
def empty_res_tree(tmpdir):
    """Empty result directory."""
    tree = Path(tmpdir) / "res"
    tree.mkdir()
    return tree


@pytest.fixture
def ref_tree(empty_ref_tree):
    """Reference directory tree."""
    generate_test_files.create_pdf(empty_ref_tree / "file.pdf")
    generate_test_files.create_json(empty_ref_tree / "file.json")
    generate_test_files.create_yaml(empty_ref_tree / "file.yaml")
    generate_test_files.create_xml(empty_ref_tree / "file.xml")
    return empty_ref_tree


@pytest.fixture
def res_tree_equal(empty_res_tree):
    """Result directory tree equal to the reference."""
    generate_test_files.create_pdf(empty_res_tree / "file.pdf")
    generate_test_files.create_json(empty_res_tree / "file.json")
    generate_test_files.create_yaml(empty_res_tree / "file.yaml")
    generate_test_files.create_xml(empty_res_tree / "file.xml")
    return empty_res_tree


@pytest.fixture
def res_tree_diff(empty_res_tree):
    """Result directory tree different from the reference."""
    generate_test_files.create_pdf(empty_res_tree / "file.pdf", diff=True)
    generate_test_files.create_json(empty_res_tree / "file.json", diff=True)
    generate_test_files.create_yaml(empty_res_tree / "file.yaml", diff=True)
    generate_test_files.create_xml(empty_res_tree / "file.xml", diff=True)
    return empty_res_tree


@pytest.fixture
def pdf_diff():
    """The diff that should be reported for the PDF files."""
    return r"The files '\S*/file.pdf' and '\S*/file.pdf' are different\."


@pytest.fixture
def dict_diff():
    """The diff that should be reported for the JSON and YAML files."""
    diff = (
        r"""The files '\S*' and '\S*' are different:\n"""
        r"""Added the value\(s\) '{"__dict_key_2__": \[1, 2, 3\], "__dict_key_3__": \[1, 2, 3\], """
        r""""__dict_key_4__": \[1, 2, 3\]}' in the '\[simple_dict\]' key\.\n"""
        r"""Added the value\(s\) '{"__nested_dict_key_2__": "nested_dict_val_2", """
        r""""nested_dict_key_1": "__nested_dict_val_1__"}' in the '\[nested_dict\]"""
        r"""\[sub_nested_dict\]' key\.\n"""
        r"""Added the value\(s\) '{"__nested_dict_key_2__": "nested_dict_val_2"}' in the """
        r"""'\[nested_list\]\[3\]\[1\]' key\.\n"""
        r"""Added the value\(s\) '{"simple_list_test": \[\["dict_key_1", \[1, 4, 3\]\], """
        r"""\["__dict_key_2__", \[1, 2, 3\]\]\]}' in the '' key\.\n"""
        r"""Changed the value of '\[int_value\]' from 1 to 2\.\n"""
        r"""Changed the value of '\[nested_dict\]\[dict_key\]\[1\]' from 2 to 4\.\n"""
        r"""Changed the value of '\[nested_dict\]\[sub_nested_dict\]\[nested_list_key\]"""
        r"""\[0\]' from 1 to 2\.\n"""
        r"""Changed the value of '\[nested_dict\]\[sub_nested_dict\]\[nested_list_key\]"""
        r"""\[1\]' from 2.5 to 2.50001\.\n"""
        r"""Changed the value of '\[nested_dict\]\[sub_nested_dict\]\[nested_list_key\]"""
        r"""\[2\]' from 'str_val' to '__str_val__'\.\n"""
        r"""Changed the value of '\[nested_list\]\[0\]' from 1 to 2\.\n"""
        r"""Changed the value of '\[nested_list\]\[1\]' from 2.5 to 2.50001\.\n"""
        r"""Changed the value of '\[nested_list\]\[2\]' from 'str_val' to '__str_val__'\.\n"""
        r"""Changed the value of '\[nested_list\]\[3\]\[0\]' from 'nested_list_val' to """
        r"""'__nested_list_val__'\.\n"""
        r"""Changed the value of '\[nested_list\]\[3\]\[1\]\[nested_dict_key_1\]' from """
        r"""'nested_dict_val_1' to '__nested_dict_val_1__'\.\n"""
        r"""Changed the value of '\[nested_list\]\[3\]\[1\]\[nested_list_key\]\[0\]' from 1 """
        r"""to 2\.\n"""
        r"""Changed the value of '\[nested_list\]\[3\]\[1\]\[nested_list_key\]\[1\]' from """
        r"""2.5 to 2.50001\.\n"""
        r"""Changed the value of '\[nested_list\]\[3\]\[1\]\[nested_list_key\]\[2\]' from """
        r"""'str_val' to '__str_val__'\.\n"""
        r"""Changed the value of '\[simple_dict\]\[dict_key_1\]\[1\]' from 2 to 4\.\n"""
        r"""Changed the value of '\[simple_list\]\[0\]' from 1 to 2\.\n"""
        r"""Changed the value of '\[simple_list\]\[1\]' from 2.5 to 2.50001\.\n"""
        r"""Changed the value of '\[simple_list\]\[2\]' from 'str_val' to '__str_val__'\.\n"""
        r"""Removed the value\(s\) '{"dict_key_2": \[1, 2, 3\]}' from '\[simple_dict\]' """
        r"""key\.\n"""
        r"""Removed the value\(s\) '{"nested_dict_key": "nested_dict_val"}' from '\["""
        r"""nested_dict\]\[sub_nested_dict\]' key\.\n"""
        r"""Removed the value\(s\) '{"nested_dict_key_2": "nested_dict_val_2"}' from """
        r"""'\[nested_list\]\[3\]\[1\]' key\.\n"""
        r"""Removed the value\(s\) '{"nested_dict_test": 0}' from '' key\."""
    )
    return diff


@pytest.fixture
def xml_diff(dict_diff):
    """The diff that should be reported for the XML files."""
    diff = dict_diff.replace("'\\[", "'\\[root\\]\\[").replace(" '' key", " '\\[root\\]' key")
    return diff


@pytest.fixture
def ref_csv(ref_tree):
    """The reference CSV file."""
    ref_data = {
        "col_a": [1, 2, 3],
        "col_b": ["a", "b", "c"],
        "col_c": [4, 5, 6],
    }
    df = pd.DataFrame(ref_data, index=["idx1", "idx2", "idx3"])
    filename = ref_tree / "file.csv"
    df.to_csv(filename, index=True, index_label="index")
    return filename


@pytest.fixture
def res_csv_equal(ref_csv, res_tree_equal):
    """The result CSV file equal to the reference."""
    df = pd.read_csv(ref_csv, index_col="index")
    filename = res_tree_equal / "file.csv"
    df.to_csv(filename, index=True, index_label="index")
    return filename


@pytest.fixture
def res_csv_diff(ref_csv, res_tree_diff):
    """The result CSV file different from the reference."""
    df = pd.read_csv(ref_csv, index_col="index")
    df.loc["idx1", "col_a"] *= 10
    df.loc["idx2", "col_b"] += "_new"
    filename = res_tree_diff / "file.csv"
    df.to_csv(filename, index=True, index_label="index")
    return filename


@pytest.fixture
def csv_diff():
    """The diff that should be reported for the CSV files."""
    return (
        r"""The files '\S*/file.csv' and '\S*/file.csv' are different:\n\n"""
        r"""Column 'col_a': Series are different\n\n"""
        r"""Series values are different \(33.33333 %\)\n"""
        r"""\[index\]: \[0, 1, 2\]\n"""
        r"""\[left\]:  \[1, 2, 3\]\n"""
        r"""\[right\]: \[10, 2, 3\]\n\n"""
        r"""Column 'col_b': Series are different\n\n"""
        r"""Series values are different \(33.33333 %\)\n"""
        r"""\[index\]: \[0, 1, 2\]\n\[left\]:  \[a, b, c\]\n\[right\]: \[a, b_new, c\]"""
    )



"""Function to create base files used for tests."""
import copy
import json
import tempfile
from pathlib import Path

import rst2pdf.createpdf
import yaml
from dicttoxml import dicttoxml

REF_DICT = {
    "int_value": 1,
    "simple_list": [1, 2.5, "str_val"],
    "nested_list": [
        1,
        2.5,
        "str_val",
        [
            "nested_list_val",
            {
                "nested_dict_key_1": "nested_dict_val_1",
                "nested_dict_key_2": "nested_dict_val_2",
                "nested_list_key": [1, 2.5, "str_val"],
            },
        ],
    ],
    "simple_dict": {
        "dict_key_1": [1, 2, 3],
        "dict_key_2": [1, 2, 3],
    },
    "nested_dict": {
        "dict_key": [1, 2, 3],
        "sub_nested_dict": {
            "nested_dict_key": "nested_dict_val",
            "nested_list_key": [1, 2.5, "str_val"],
        },
    },
    "nested_dict_test": 0,
}

DIFF_DICT = {
    "int_value": 2,
    "simple_list": [2, 2.50001, "__str_val__"],
    "nested_list": [
        2,
        2.50001,
        "__str_val__",
        [
            "__nested_list_val__",
            {
                "nested_dict_key_1": "__nested_dict_val_1__",
                "__nested_dict_key_2__": "nested_dict_val_2",
                "nested_list_key": [2, 2.50001, "__str_val__"],
            },
        ],
    ],
    "simple_dict": {
        "dict_key_1": [1, 4, 3],
        "__dict_key_2__": [1, 2, 3],
        "__dict_key_3__": [1, 2, 3],
        "__dict_key_4__": [1, 2, 3],
    },
    "simple_list_test": [
        ("dict_key_1", [1, 4, 3]),
        ("__dict_key_2__", [1, 2, 3]),
    ],
    "nested_dict": {
        "dict_key": [1, 4, 3],
        "sub_nested_dict": {
            "nested_dict_key_1": "__nested_dict_val_1__",
            "__nested_dict_key_2__": "nested_dict_val_2",
            "nested_list_key": [2, 2.50001, "__str_val__"],
        },
    },
}

REF_RST = r"""
This is a test file
===================

This is a random sentence.

This is a simple math formula:

.. math::

    x = 3 \times y^2

"""

DIFF_RST = r"""
This is a test file
===================

This is **a** random sentence.

This is a simple math formula:

.. math::

    x = 8 \times y^2

"""


def create_json(filename, diff=False):
    """Create a JSON file."""
    if diff:
        data = copy.deepcopy(DIFF_DICT)
    else:
        data = copy.deepcopy(REF_DICT)
    with open(filename, "w", encoding="utf-8") as f:
        json.dump(data, f)


def create_yaml(filename, diff=False):
    """Create a YAML file."""
    if diff:
        data = copy.deepcopy(DIFF_DICT)
    else:
        data = copy.deepcopy(REF_DICT)
    with open(filename, "w", encoding="utf-8") as f:
        yaml.dump(data, f)


def create_xml(filename, diff=False):
    """Create a YAML file."""
    if diff:
        data = copy.deepcopy(DIFF_DICT)
    else:
        data = copy.deepcopy(REF_DICT)

    with open(filename, "w", encoding="utf-8") as f:
        xml_data = dicttoxml(data).decode("utf-8")
        # Remove a type attribute for test purpose
        xml_data = xml_data.replace('nested_dict_key_1 type="str"', "nested_dict_key_1")
        f.write(xml_data)


def create_pdf(filename, diff=False):
    """Create a PDF file."""
    if diff:
        data = copy.deepcopy(DIFF_RST)
    else:
        data = copy.deepcopy(REF_RST)
    with tempfile.TemporaryDirectory() as tmp_dir:
        rst_file = Path(tmp_dir) / Path(filename.name).with_suffix(".rst")
        with open(rst_file, "w", encoding="utf-8") as f:
            f.write(data)

        try:
            rst2pdf.createpdf.main([str(rst_file), "-o", str(filename)])
        except SystemExit as exc:
            if exc.code != 0:
                raise exc



"""Test the base features of the ``dir-content-diff`` package."""
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=unused-argument
# pylint: disable=use-implicit-booleaness-not-comparison
import json
import re

import dictdiffer
import pytest

import dir_content_diff
from dir_content_diff import assert_equal_trees
from dir_content_diff import compare_trees


class TestBaseComparator:
    """Test the base comparator."""

    def test_equal(self):
        """Test equal."""
        assert dir_content_diff.JsonComparator() != dir_content_diff.PdfComparator()
        assert dir_content_diff.JsonComparator() == dir_content_diff.JsonComparator()

        class ComparatorWithAttributes(dir_content_diff.base_comparators.BaseComparator):
            """Compare data from two JSON files."""

            def __init__(self, arg1, arg2):
                super().__init__()
                self.arg1 = arg1
                if arg2:
                    self.arg2 = arg2

            def diff(self, ref, comp, *args, **kwargs):
                return False

        assert ComparatorWithAttributes(1, 2) == ComparatorWithAttributes(1, 2)
        assert ComparatorWithAttributes(1, 2) != ComparatorWithAttributes(3, 4)
        assert ComparatorWithAttributes(1, 2) != ComparatorWithAttributes(1, None)

    def test_load_kwargs(self, ref_tree, res_tree_diff):
        """Test the load_kwargs method."""

        class ComparatorWithLoader(dir_content_diff.base_comparators.JsonComparator):
            """Compare data from two JSON files."""

            def load(self, path, load_empty=False):
                if load_empty:
                    return {}
                return super().load(path)

        ref_file = ref_tree / "file.json"
        res_file = res_tree_diff / "file.json"

        diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithLoader(),
        )

        no_load_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithLoader(),
            load_kwargs={"load_empty": False},
        )

        no_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithLoader(),
            load_kwargs={"load_empty": True},
        )

        no_diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithLoader(default_load_kwargs={"load_empty": True}),
        )

        diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithLoader(default_load_kwargs={"load_empty": True}),
            load_kwargs={"load_empty": False},
        )

        kwargs_msg = "Kwargs used for loading data: {'load_empty': False}\n"
        assert kwargs_msg in no_load_diff
        assert diff == no_load_diff.replace(kwargs_msg, "")
        assert diff is not False
        assert no_diff is False
        assert no_diff_default is False
        assert kwargs_msg in diff_default
        assert diff_default.replace(kwargs_msg, "") == diff

    def test_filter_kwargs(self, ref_tree, res_tree_diff):
        """Test the filter_kwargs method."""

        class ComparatorWithFilter(dir_content_diff.base_comparators.JsonComparator):
            """Compare data from two JSON files."""

            def filter(self, differences, remove_all=False):
                if remove_all:
                    return []
                return differences

        ref_file = ref_tree / "file.json"
        res_file = res_tree_diff / "file.json"

        diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithFilter(),
        )

        no_filter_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithFilter(),
            filter_kwargs={"remove_all": False},
        )

        no_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithFilter(),
            filter_kwargs={"remove_all": True},
        )

        no_diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithFilter(default_filter_kwargs={"remove_all": True}),
        )

        diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithFilter(default_filter_kwargs={"remove_all": True}),
            filter_kwargs={"remove_all": False},
        )

        kwargs_msg = "Kwargs used for filtering differences: {'remove_all': False}\n"
        assert kwargs_msg in no_filter_diff
        assert diff == no_filter_diff.replace(kwargs_msg, "")
        assert diff is not False
        assert no_diff is False
        assert no_diff_default is False
        assert kwargs_msg in diff_default
        assert diff_default.replace(kwargs_msg, "") == diff

    def test_format_kwargs(self, ref_tree, res_tree_diff):
        """Test the format_kwargs method."""

        class ComparatorWithFormat(dir_content_diff.base_comparators.JsonComparator):
            """Compare data from two JSON files."""

            def format_diff(self, difference, mark_formatted=False):
                """Format one element difference."""
                difference = super().format_diff(difference)
                if mark_formatted:
                    difference += "### FORMATTED"
                return difference

        ref_file = ref_tree / "file.json"
        res_file = res_tree_diff / "file.json"

        diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithFormat(),
        )

        no_format_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithFormat(),
            format_diff_kwargs={"mark_formatted": False},
        )

        formatted_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithFormat(),
            format_diff_kwargs={"mark_formatted": True},
        )

        formatted_diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithFormat(default_format_diff_kwargs={"mark_formatted": True}),
        )

        diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithFormat(default_format_diff_kwargs={"mark_formatted": True}),
            format_diff_kwargs={"mark_formatted": False},
        )

        kwargs_msg = "Kwargs used for formatting differences: {'mark_formatted': False}\n"
        assert kwargs_msg in no_format_diff
        assert diff == no_format_diff.replace(kwargs_msg, "")
        assert len(re.findall("### FORMATTED", diff)) == 0
        assert len(re.findall("### FORMATTED", formatted_diff)) == 25
        assert len(re.findall("### FORMATTED", formatted_diff_default)) == 25
        assert kwargs_msg in diff_default
        assert diff_default.replace(kwargs_msg, "") == diff

    def test_sort_kwargs(self, ref_tree, res_tree_diff):
        """Test the sort_kwargs method."""

        class ComparatorWithSort(dir_content_diff.base_comparators.JsonComparator):
            """Compare data from two JSON files."""

            def sort(self, differences, reverse=False):
                """Sort the element differences."""
                return sorted(differences, reverse=reverse)

        ref_file = ref_tree / "file.json"
        res_file = res_tree_diff / "file.json"

        diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithSort(),
        )

        no_reversed_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithSort(),
            sort_kwargs={"reverse": False},
        )

        reversed_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithSort(),
            sort_kwargs={"reverse": True},
        )

        reversed_diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithSort(default_sort_kwargs={"reverse": True}),
        )

        no_reversed_diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithSort(default_sort_kwargs={"reverse": True}),
            sort_kwargs={"reverse": False},
        )

        kwargs_msg = "Kwargs used for sorting differences: {'reverse': True}\n"
        kwargs_msg_false = kwargs_msg.replace("True", "False")
        expected_reversed_diff = "\n".join(
            diff.split("\n")[:1] + sorted(diff.split("\n")[1:], reverse=True)
        )

        assert kwargs_msg not in diff
        assert kwargs_msg_false not in diff
        assert kwargs_msg_false in no_reversed_diff
        assert kwargs_msg in reversed_diff
        assert kwargs_msg in reversed_diff_default
        assert kwargs_msg_false in no_reversed_diff_default

        assert diff == no_reversed_diff.replace(kwargs_msg_false, "")
        assert expected_reversed_diff == reversed_diff.replace(kwargs_msg, "")
        assert expected_reversed_diff == reversed_diff_default.replace(kwargs_msg, "")
        assert diff == no_reversed_diff_default.replace(kwargs_msg_false, "")

    def test_concat_kwargs(self, ref_tree, res_tree_diff):
        """Test the concat_kwargs method."""

        class ComparatorWithConcat(dir_content_diff.base_comparators.JsonComparator):
            """Compare data from two JSON files."""

            def concatenate(self, differences, eol=None):
                """Concatenate the differences."""
                if not eol:
                    eol = "\n"
                return eol.join(differences)

        ref_file = ref_tree / "file.json"
        res_file = res_tree_diff / "file.json"

        diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithConcat(),
        )

        concat_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithConcat(),
            concat_kwargs={"eol": "\n"},
        )

        concat_eol_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithConcat(),
            concat_kwargs={"eol": "#EOL#"},
        )

        concat_eol_diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithConcat(default_concat_kwargs={"eol": "#EOL#"}),
        )

        concat_diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithConcat(default_concat_kwargs={"eol": "#EOL#"}),
            concat_kwargs={"eol": "\n"},
        )

        kwargs_msg_eol = "\nKwargs used for concatenating differences: {'eol': '#EOL#'}\n"
        kwargs_msg_n = kwargs_msg_eol.replace("#EOL#", "\\n")
        TEST_EOL = "__TEST_EOL__"

        assert kwargs_msg_eol not in diff
        assert kwargs_msg_n not in diff
        assert kwargs_msg_n in concat_diff
        assert kwargs_msg_eol in concat_eol_diff
        assert kwargs_msg_eol in concat_eol_diff_default
        assert kwargs_msg_n in concat_diff_default

        assert diff == concat_diff.replace(kwargs_msg_n, "\n")
        assert concat_diff.replace(kwargs_msg_n, "").replace(
            "\n", TEST_EOL
        ) == concat_eol_diff.replace(kwargs_msg_eol, "").replace("#EOL#", TEST_EOL)
        assert concat_eol_diff == concat_eol_diff_default
        assert diff == concat_diff_default.replace(kwargs_msg_n, "\n")

    def test_report_kwargs(self, ref_tree, res_tree_diff):
        """Test the report_kwargs method."""

        class ComparatorWithReport(dir_content_diff.base_comparators.JsonComparator):
            """Compare data from two JSON files."""

            def report(
                self,
                ref_file,
                comp_file,
                formatted_differences,
                diff_args,
                diff_kwargs,
                mark_report=None,
                **kwargs,
            ):
                if mark_report is not None:
                    kwargs["mark_report"] = mark_report
                report = super().report(
                    ref_file, comp_file, formatted_differences, diff_args, diff_kwargs, **kwargs
                )
                if mark_report:
                    report += "### REPORTED"
                return report

        ref_file = ref_tree / "file.json"
        res_file = res_tree_diff / "file.json"

        diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithReport(),
        )

        no_report_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithReport(),
            report_kwargs={"mark_report": False},
        )

        reported_diff = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithReport(),
            report_kwargs={"mark_report": True},
        )

        reported_diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithReport(default_report_kwargs={"mark_report": True}),
        )

        no_report_diff_default = dir_content_diff.compare_files(
            ref_file,
            res_file,
            ComparatorWithReport(default_report_kwargs={"mark_report": True}),
            report_kwargs={"mark_report": False},
        )

        kwargs_msg = "Kwargs used for reporting differences: {'mark_report': False}\n"
        assert kwargs_msg in no_report_diff
        assert diff == no_report_diff.replace(kwargs_msg, "")
        assert len(re.findall("### REPORTED", diff)) == 0
        assert len(re.findall("### REPORTED", reported_diff)) == 1
        assert len(re.findall("### REPORTED", reported_diff_default)) == 1
        assert kwargs_msg in no_report_diff_default
        assert no_report_diff_default.replace(kwargs_msg, "") == diff

    class TestXmlComparator:
        """Test the XML comparator."""

        def test_xmltodict(self):
            """Test all types of the xmltodict auto cast feature."""
            comparator = dir_content_diff.XmlComparator()

            # Test empty root
            res = comparator.xmltodict(
                # fmt: off
                """<?xml version="1.0" encoding="UTF-8" ?>"""
                """<root>"""
                """</root>"""
                # fmt: on
            )
            assert res == {"root": {}}

            # Test all types
            res = comparator.xmltodict(
                """<?xml version="1.0" encoding="UTF-8" ?>"""
                """<root>"""
                """    <str_value_no_type>a str value</str_value_no_type>"""
                """    <str_value type="str">another str value</str_value>"""
                """    <int_value type="int">1</int_value>"""
                """    <float_value type="float">1.5</float_value>"""
                """    <boolean_true type="bool">TrUe</boolean_true>"""
                """    <boolean_false type="bool">FaLsE</boolean_false>"""
                """    <simple_list type="list">"""
                """        <item type="int">1</item>"""
                """        <item type="float">2.5</item>"""
                """        <item type="str">str_val</item>"""
                """    </simple_list>"""
                """    <simple_dict type="dict">"""
                """        <key_1 type="int">1</key_1>"""
                """        <key_2 type="float">2.5</key_2>"""
                """        <key_3 type="str">str_val</key_3>"""
                """        <key_4>another str val</key_4>"""
                """    </simple_dict>"""
                """    <none type="null">any thing here is not considered</none>"""
                """</root>"""
            )
            assert res["root"]["none"] is None
            del res["root"]["none"]
            assert res == {
                "root": {
                    "str_value_no_type": "a str value",
                    "str_value": "another str value",
                    "int_value": 1,
                    "float_value": 1.5,
                    "boolean_true": True,
                    "boolean_false": False,
                    "simple_list": [1, 2.5, "str_val"],
                    "simple_dict": {
                        "key_1": 1,
                        "key_2": 2.5,
                        "key_3": "str_val",
                        "key_4": "another str val",
                    },
                }
            }

            # Test unknown type
            with pytest.raises(TypeError, match=r"Unsupported type.*"):
                comparator.xmltodict(
                    """<?xml version="1.0" encoding="UTF-8" ?>"""
                    """<root>"""
                    """    <bad_type type="UNKNOWN TYPE">1</bad_type>"""
                    """</root>"""
                )

            # Test bad value in boolean
            with pytest.raises(ValueError, match="Bool attributes expect 'true' or 'false'."):
                comparator.xmltodict(
                    """<?xml version="1.0" encoding="UTF-8" ?>"""
                    """<root>"""
                    """    <bad_bool type="bool">not a bool</bad_bool>"""
                    """</root>"""
                )

        def test_add_to_output_with_none(self):
            """Test wrong type for add_to_output() method."""
            comparator = dir_content_diff.XmlComparator()
            comparator.add_to_output(None, None)


class TestRegistry:
    """Test the internal registry."""

    def test_init_register(self, registry_reseter):
        """Test the initial registry with the get_comparators() function."""
        assert dir_content_diff.get_comparators() == {
            None: dir_content_diff.DefaultComparator(),
            ".json": dir_content_diff.JsonComparator(),
            ".pdf": dir_content_diff.PdfComparator(),
            ".yaml": dir_content_diff.YamlComparator(),
            ".yml": dir_content_diff.YamlComparator(),
            ".xml": dir_content_diff.XmlComparator(),
        }

    def test_update_register(self, registry_reseter):
        """Test the functions to update the registry."""
        dir_content_diff.register_comparator(".test_ext", dir_content_diff.JsonComparator())
        assert dir_content_diff.get_comparators() == {
            None: dir_content_diff.DefaultComparator(),
            ".test_ext": dir_content_diff.JsonComparator(),
            ".json": dir_content_diff.JsonComparator(),
            ".pdf": dir_content_diff.PdfComparator(),
            ".yaml": dir_content_diff.YamlComparator(),
            ".yml": dir_content_diff.YamlComparator(),
            ".xml": dir_content_diff.XmlComparator(),
        }

        dir_content_diff.unregister_comparator(".yaml")
        dir_content_diff.unregister_comparator("json")  # Test suffix without dot
        assert dir_content_diff.get_comparators() == {
            None: dir_content_diff.DefaultComparator(),
            ".test_ext": dir_content_diff.JsonComparator(),
            ".pdf": dir_content_diff.PdfComparator(),
            ".yml": dir_content_diff.YamlComparator(),
            ".xml": dir_content_diff.XmlComparator(),
        }

        dir_content_diff.reset_comparators()
        assert dir_content_diff.get_comparators() == {
            None: dir_content_diff.DefaultComparator(),
            ".json": dir_content_diff.JsonComparator(),
            ".pdf": dir_content_diff.PdfComparator(),
            ".yaml": dir_content_diff.YamlComparator(),
            ".yml": dir_content_diff.YamlComparator(),
            ".xml": dir_content_diff.XmlComparator(),
        }

        with pytest.raises(
            ValueError,
            match=(
                "The '.pdf' extension is already registered and must be unregistered before being "
                "replaced."
            ),
        ):
            dir_content_diff.register_comparator(".pdf", dir_content_diff.JsonComparator())

        with pytest.raises(ValueError, match=("The '.unknown_ext' extension is not registered.")):
            dir_content_diff.unregister_comparator(".unknown_ext")

        dir_content_diff.unregister_comparator(".unknown_ext", quiet=True)
        dir_content_diff.register_comparator(".new_ext", dir_content_diff.JsonComparator())
        assert dir_content_diff.get_comparators() == {
            None: dir_content_diff.DefaultComparator(),
            ".json": dir_content_diff.JsonComparator(),
            ".pdf": dir_content_diff.PdfComparator(),
            ".yaml": dir_content_diff.YamlComparator(),
            ".yml": dir_content_diff.YamlComparator(),
            ".xml": dir_content_diff.XmlComparator(),
            ".new_ext": dir_content_diff.JsonComparator(),
        }
        dir_content_diff.register_comparator(
            ".new_ext", dir_content_diff.PdfComparator(), force=True
        )
        assert dir_content_diff.get_comparators() == {
            None: dir_content_diff.DefaultComparator(),
            ".json": dir_content_diff.JsonComparator(),
            ".pdf": dir_content_diff.PdfComparator(),
            ".yaml": dir_content_diff.YamlComparator(),
            ".yml": dir_content_diff.YamlComparator(),
            ".xml": dir_content_diff.XmlComparator(),
            ".new_ext": dir_content_diff.PdfComparator(),
        }


@pytest.fixture
def ref_with_nested_file(ref_tree):
    """Update the ref tree to have nesteed files."""
    ref_pdf_file = ref_tree / "file.pdf"
    new_ref_pdf_file = ref_tree / "level1" / "level2" / "level3" / "file.pdf"
    new_ref_pdf_file.parent.mkdir(parents=True)
    ref_pdf_file.rename(new_ref_pdf_file)
    return ref_tree


@pytest.fixture
def res_equal_with_nested_file(res_tree_equal):
    """Update the result tree to have nesteed files."""
    res_pdf_file = res_tree_equal / "file.pdf"
    new_res_pdf_file = res_tree_equal / "level1" / "level2" / "level3" / "file.pdf"
    new_res_pdf_file.parent.mkdir(parents=True)
    res_pdf_file.rename(new_res_pdf_file)
    return res_tree_equal


@pytest.fixture
def res_diff_with_nested_file(res_tree_diff):
    """Update the result tree to have nesteed files."""
    res_pdf_file = res_tree_diff / "file.pdf"
    new_res_pdf_file = res_tree_diff / "level1" / "level2" / "level3" / "file.pdf"
    new_res_pdf_file.parent.mkdir(parents=True)
    res_pdf_file.rename(new_res_pdf_file)
    return res_tree_diff


class TestEqualTrees:
    """Tests that should return no difference."""

    def test_diff_tree(self, ref_tree, res_tree_equal):
        """Test that no difference is returned."""
        res = compare_trees(ref_tree, res_tree_equal)
        assert res == {}

    def test_assert_equal_trees(self, ref_tree, res_tree_equal):
        """Test that no exception is raised."""
        assert_equal_trees(ref_tree, res_tree_equal)

    def test_assert_equal_trees_export(self, ref_tree, res_tree_equal):
        """Test that the formatted files are properly exported."""

        class JsonComparator(dir_content_diff.base_comparators.JsonComparator):
            """Compare data from two JSON files."""

            def save(self, data, path):
                """Save formatted data into a file."""
                with open(path, "w", encoding="utf-8") as file:
                    json.dump(data, file)

        comparators = dir_content_diff.get_comparators()
        comparators[".json"] = JsonComparator()
        assert_equal_trees(
            ref_tree, res_tree_equal, export_formatted_files=True, comparators=comparators
        )
        assert list(res_tree_equal.with_name(res_tree_equal.name + "_FORMATTED").iterdir()) == [
            res_tree_equal.with_name(res_tree_equal.name + "_FORMATTED") / "file.json"
        ]

    def test_diff_empty(self, empty_ref_tree, empty_res_tree):
        """Test with empty trees."""
        res = compare_trees(empty_ref_tree, empty_res_tree)
        assert res == {}

    def test_pass_register(self, empty_ref_tree, empty_res_tree):
        """Test with empty trees and with an explicit set of comparators."""
        res = compare_trees(
            empty_ref_tree, empty_res_tree, comparators=dir_content_diff.get_comparators()
        )
        assert res == {}

    def test_unknown_comparator(self, ref_tree, res_tree_equal, registry_reseter):
        """Test with an unknown extension."""
        dir_content_diff.unregister_comparator(".yaml")
        res = compare_trees(ref_tree, res_tree_equal)
        assert res == {}

    def test_nested_files(self, ref_with_nested_file, res_equal_with_nested_file):
        """Test with nested files."""
        res = compare_trees(ref_with_nested_file, res_equal_with_nested_file)
        assert res == {}

    def test_specific_args(self, ref_tree, res_tree_equal):
        """Test specific args."""
        specific_args = {
            "file.yaml": {"args": [None, None, None, False, 0, False]},
            "file.json": {"tolerance": 0},
        }
        res = compare_trees(ref_tree, res_tree_equal, specific_args=specific_args)

        assert res == {}


class TestDiffTrees:
    """Tests that should return differences."""

    def test_diff_tree(self, ref_tree, res_tree_diff, pdf_diff, dict_diff, xml_diff):
        """Test that the returned differences are correct."""
        res = compare_trees(ref_tree, res_tree_diff)

        assert len(res) == 4
        match_res_0 = re.match(pdf_diff, res["file.pdf"])
        match_res_1 = re.match(dict_diff, res["file.json"])
        match_res_2 = re.match(dict_diff, res["file.yaml"])
        match_res_3 = re.match(xml_diff, res["file.xml"])

        for match_i in [match_res_0, match_res_1, match_res_2, match_res_3]:
            assert match_i is not None

    def test_assert_equal_trees(self, ref_tree, res_tree_diff, pdf_diff, dict_diff, xml_diff):
        """Test that the exception raised is correct."""
        pattern = (r"\n\n\n").join([dict_diff, pdf_diff, xml_diff, dict_diff])
        with pytest.raises(AssertionError, match=pattern):
            assert_equal_trees(ref_tree, res_tree_diff)

    def test_diff_ref_empty_res_not_empty(self, empty_ref_tree, res_tree_equal):
        """Test with empty ref tree."""
        res = compare_trees(empty_ref_tree, res_tree_equal)
        assert res == {}

    def test_diff_ref_not_empty_res_empty(self, ref_tree, empty_res_tree):
        """Test with empty compared tree."""
        res = compare_trees(ref_tree, empty_res_tree)

        assert len(res) == 4
        match_res_0 = re.match(
            r"The file 'file.pdf' does not exist in '\S*/res'\.", res["file.pdf"]
        )
        match_res_1 = re.match(
            r"The file 'file.yaml' does not exist in '\S*/res'\.", res["file.yaml"]
        )
        match_res_2 = re.match(
            r"The file 'file.json' does not exist in '\S*/res'\.", res["file.json"]
        )
        match_res_3 = re.match(
            r"The file 'file.xml' does not exist in '\S*/res'\.", res["file.xml"]
        )

        for match_i in [match_res_0, match_res_1, match_res_2, match_res_3]:
            assert match_i is not None

    def test_exception_in_comparator(self, ref_tree, res_tree_equal, registry_reseter):
        """Test with a comparator raising an exception."""

        def bad_comparator(ref_path, test_path, *args, **kwargs):
            raise RuntimeError("Bad\ncomparator")

        dir_content_diff.unregister_comparator(".yaml")
        dir_content_diff.register_comparator(".yaml", bad_comparator)

        res = compare_trees(ref_tree, res_tree_equal)

        assert len(res) == 1
        match = re.match(
            r"The files '\S*/ref/file.yaml' and '\S*/res/file.yaml' are different:\n"
            r"Exception raised: Bad\ncomparator",
            res["file.yaml"],
        )
        assert match is not None

    def test_specific_args(self, ref_tree, res_tree_diff, dict_diff, xml_diff):
        """Test specific args."""
        specific_args = {
            "file.pdf": {"threshold": 50},
            "file.json": {"tolerance": 0},
        }
        res = compare_trees(ref_tree, res_tree_diff, specific_args=specific_args)

        # This time the PDF files are considered as equal
        assert len(res) == 3
        match_res_0 = re.match(dict_diff, res["file.yaml"])
        match_res_1 = re.match(
            dict_diff.replace(
                r"are different:\n",
                r"are different:\nKwargs used for computing differences: \{'tolerance': 0\}\n",
            ),
            res["file.json"],
        )
        match_res_2 = re.match(xml_diff, res["file.xml"])

        for match_i in [match_res_0, match_res_1, match_res_2]:
            assert match_i is not None

    def test_unknown_comparator(self, ref_tree, res_tree_diff, registry_reseter):
        """Test with an unknown extension."""
        dir_content_diff.unregister_comparator(".yaml")
        res = compare_trees(ref_tree, res_tree_diff)
        match = re.match(
            r"The files '\S*/ref/file.yaml' and '\S*/res/file.yaml' are different.",
            res["file.yaml"],
        )
        assert match is not None

    def test_nested_files(self, ref_with_nested_file, res_diff_with_nested_file):
        """Test with nested files."""
        res = compare_trees(ref_with_nested_file, res_diff_with_nested_file)
        match = re.match(
            r"The files '\S*/ref/level1/level2/level3/file.pdf' and "
            r"'\S*/res/level1/level2/level3/file.pdf' are different\.",
            res["level1/level2/level3/file.pdf"],
        )
        assert match is not None

    def test_fix_dot_notation(self, ref_tree, res_tree_diff, pdf_diff, dict_diff, xml_diff):
        """Test that the dot notation is properly fixed."""
        specific_args = {"file.yaml": {"args": [None, None, None, False, 0, True]}}
        res = compare_trees(ref_tree, res_tree_diff, specific_args=specific_args)

        assert len(res) == 4
        match_res_0 = re.match(pdf_diff, res["file.pdf"])
        match_res_1 = re.match(
            dict_diff.replace(
                r"are different:\n",
                r"are different:\nArgs used for computing differences: "
                r"\[None, None, None, False, 0, True\]\n",
            ),
            res["file.yaml"],
        )
        match_res_2 = re.match(dict_diff, res["file.json"])
        match_res_3 = re.match(xml_diff, res["file.xml"])

        for match_i in [match_res_0, match_res_1, match_res_2, match_res_3]:
            assert match_i is not None

    def test_format_inside_diff(self, ref_tree, res_tree_diff, dict_diff):
        """Test formatting the result inside the diff method."""

        class JsonComparator(dir_content_diff.base_comparators.BaseComparator):
            """Compare data from two JSON files."""

            def load(self, path, *args, **kwargs):
                with open(path) as file:  # pylint: disable=unspecified-encoding
                    data = json.load(file)
                return data

            def diff(self, ref, comp, *args, **kwargs):
                diffs = list(dictdiffer.diff(ref, comp, *args, dot_notation=False, **kwargs))

                # Format here instead of overriding the default format method
                comparator = dir_content_diff.base_comparators.JsonComparator()
                formatted = [comparator.format_diff(i) for i in diffs]

                return formatted

        res = compare_trees(ref_tree, res_tree_diff, comparators={".json": JsonComparator()})

        match = re.match(dict_diff, res["file.json"])

        assert match is not None


class TestProgrammaticUse:
    """Test specific comparators that could be use programmatically."""

    def test_diff_tree(self, ref_tree, res_tree_diff, pdf_diff, dict_diff):
        """Test with different trees."""
        res = compare_trees(ref_tree, res_tree_diff, return_raw_diffs=True)

        res_json = res["file.json"]

        assert len(res_json) == 25
        assert len(list(filter(lambda x: x[0] == "change", res_json))) == 17
        assert len(list(filter(lambda x: x[0] == "add", res_json))) == 4
        assert len(list(filter(lambda x: x[0] == "remove", res_json))) == 4



"""Plugin for pytest.

Register this package as a pytest plugin. When run with pytest, the
:func:`dir_content_diff.assert_equal_trees` function can use the following command line options of
pytest:

* ``--dcd-export-formatted-data`` to trigger the export of the formatted files.
* ``--dcd-export-suffix`` to specify a custom suffix (the default is ``_FORMATTED``).

When this feature is triggered, the files that are processed by a comparator with a ``save``
capability will be formatted and exported to a directory whose name is built from the compared
directory with a suffix added. The default suffix is ``_FORMATTED`` and can be customized.
"""
from dir_content_diff import _DEFAULT_EXPORT_SUFFIX
from dir_content_diff import assert_equal_trees


def pytest_addoption(parser):
    """Add command line options for pytest."""
    group = parser.getgroup("dir-content-diff", "dir-content-diff integration")
    group.addoption(
        "--dcd-export-formatted-data",
        action="store_true",
        default=False,
        help=(
            "Format the data and export it to a new directory whose name is the reference "
            "directory name and a given suffix."
        ),
    )
    group.addoption(
        "--dcd-export-suffix",
        default=_DEFAULT_EXPORT_SUFFIX,
        help="Suffix added to the reference directory.",
    )


def pytest_configure(config):
    """Process cmdline arguments."""
    # pylint: disable=protected-access
    assert_equal_trees._pytest_export_formatted_data = config.getoption(
        "--dcd-export-formatted-data"
    )
    assert_equal_trees._pytest_export_suffix = config.getoption("--dcd-export-suffix")



"""Extension module to process files with Pandas."""
try:
    import pandas as pd
except ImportError as exception:  # pragma: no cover
    raise ImportError("Could not import pandas package, please install it.") from exception

from dir_content_diff import register_comparator
from dir_content_diff.base_comparators import BaseComparator


class DataframeComparator(BaseComparator):
    """Comparator for :class:`pandas.DataFrame` objects."""

    def format_data(self, data, ref=None, replace_pattern=None):
        """Format the compared :class:`pandas.DataFrame`.

        Args:
            data (pandas.DataFrame): The DataFrame to format.
            ref (pandas.DataFrame): (Optional) The reference DataFrame.
            **replace_pattern (dict): (Optional) The columns that contain a given pattern which
                must be made replaced.
                The dictionary must have the following format:

                .. code-block:: python

                    {
                        (<pattern>, <new_value>, <optional regex flag>): [col1, col2]
                    }

        .. note::
            The formatting errors are stored in `self.current_state["format_errors"]`.
            It contains a dict in which the keys are the columns with detected issues and the
            values are the actual descriptions of these issues.

        Returns:
            pandas.DataFrame: The formatted compared data.
        """
        self.current_state["format_errors"] = errors = {}

        if replace_pattern is not None:
            for pat, cols in replace_pattern.items():
                pattern = pat[0]
                new_value = pat[1]
                if len(pat) > 2:
                    flags = pat[2]
                else:
                    flags = 0
                for col in cols:
                    if ref is not None and col not in ref.columns:
                        errors[col] = (
                            "The column is missing in the reference DataFrame, please fix the "
                            "'replace_pattern' argument."
                        )
                    elif col not in data.columns:
                        errors[col] = (
                            "The column is missing in the compared DataFrame, please fix the "
                            "'replace_pattern' argument."
                        )
                    elif hasattr(data[col], "str"):
                        # If all values are NaN, Pandas casts the column dtype to float, so the str
                        # attribute is not available.
                        data[col] = data[col].str.replace(
                            pattern,
                            new_value,
                            flags=flags,
                            regex=True,
                        )

        return data

    def diff(self, ref, comp, *args, ignore_columns=None, **kwargs):
        """Compare two :class:`pandas.DataFrame` objects.

        This function calls :func:`pandas.testing.assert_series_equal`, read the doc of this
        function for details on args and kwargs.

        Args:
            ref (pandas.DataFrame): The reference DataFrame.
            comp (pandas.DataFrame): The compared DataFrame.
            **ignore_columns (list(str)): (Optional) The columns that should not be checked.

        Returns:
            bool or str: ``False`` if the DataFrames are considered as equal or a string explaining
            why they are not considered equal.
        """
        errors = self.current_state.get("format_errors", {})

        if ignore_columns is not None:
            ref.drop(columns=ignore_columns, inplace=True, errors="ignore")
            comp.drop(columns=ignore_columns, inplace=True, errors="ignore")

        for col in ref.columns:
            if col in errors:
                continue
            try:
                if col not in comp.columns:
                    errors[col] = "The column is missing in the compared DataFrame."
                else:
                    pd.testing.assert_series_equal(ref[col], comp[col], *args, **kwargs)
                    errors[col] = True
            except AssertionError as e:
                errors[col] = e.args[0]

        for col in comp.columns:
            if col not in errors and col not in ref.columns:
                errors[col] = "The column is missing in the reference DataFrame."

        not_equals = {k: v for k, v in errors.items() if v is not True}
        if len(not_equals) == 0:
            return False
        return not_equals

    def format_diff(self, difference):
        """Format one element difference."""
        k, v = difference
        return f"\nColumn '{k}': {v}"

    def sort(self, differences):
        """Do not sort the differences to keep the column order."""
        return differences


class CsvComparator(DataframeComparator):
    """Comparator for CSV files."""

    def load(self, path, **kwargs):
        """Load a CSV file into a :class:`pandas.DataFrame` object."""
        return pd.read_csv(path, **kwargs)

    def save(self, data, path, **kwargs):
        """Save data to a CSV file."""
        index = kwargs.pop("index", False)
        data.to_csv(path, index=index, **kwargs)


class HdfComparator(DataframeComparator):
    """Comparator for HDF files."""

    def load(self, path, **kwargs):
        """Load a HDF file into a :class:`pandas.DataFrame` object."""
        return pd.read_hdf(path, **kwargs)

    def save(self, data, path, **kwargs):
        """Save data to a HDF file."""
        index = kwargs.pop("index", False)
        key = kwargs.pop("key", "data")
        data.to_hdf(path, index=index, key=key, **kwargs)


def register():
    """Register Pandas extensions."""
    register_comparator(".csv", CsvComparator())
    register_comparator(".tsv", CsvComparator())
    register_comparator(".h4", HdfComparator())
    register_comparator(".h5", HdfComparator())
    register_comparator(".hdf", HdfComparator())
    register_comparator(".hdf4", HdfComparator())
    register_comparator(".hdf5", HdfComparator())



"""Module containing the base comparators."""
import filecmp
import json
from abc import ABC
from abc import abstractmethod
from xml.etree import ElementTree

import dictdiffer
import yaml
from diff_pdf_visually import pdfdiff

from dir_content_diff.util import diff_msg_formatter

_ACTION_MAPPING = {
    "add": "Added the value(s) '{value}' in the '{key}' key.",
    "change": "Changed the value of '{key}' from {value[0]} to {value[1]}.",
    "remove": "Removed the value(s) '{value}' from '{key}' key.",
}


class BaseComparator(ABC):
    """Base Comparator class."""

    def __init__(
        self,
        default_load_kwargs=None,
        default_format_data_kwargs=None,
        default_diff_kwargs=None,
        default_filter_kwargs=None,
        default_format_diff_kwargs=None,
        default_sort_kwargs=None,
        default_concat_kwargs=None,
        default_report_kwargs=None,
        default_save_kwargs=None,
    ):
        self._default_load_kwargs = default_load_kwargs or {}
        self._default_format_data_kwargs = default_format_data_kwargs or {}
        self._default_diff_kwargs = default_diff_kwargs or {}
        self._default_filter_kwargs = default_filter_kwargs or {}
        self._default_format_diff_kwargs = default_format_diff_kwargs or {}
        self._default_sort_kwargs = default_sort_kwargs or {}
        self._default_concat_kwargs = default_concat_kwargs or {}
        self._default_report_kwargs = default_report_kwargs or {}
        self._default_save_kwargs = default_save_kwargs or {}

        self.current_state = {}

    def load(self, path, **kwargs):
        """Load a file."""
        return path

    def format_data(self, data, ref=None, **kwargs):
        """Format the loaded data."""
        # pylint: disable=unused-argument
        return data

    def save(self, data, path, **kwargs):
        """Save formatted data into a file."""
        raise NotImplementedError  # pragma: no cover

    @property
    def save_capability(self):
        """Check that the current class has a ``save()`` capability."""
        return self.__class__.save != BaseComparator.save

    @abstractmethod
    def diff(self, ref, comp, *args, **kwargs):
        """Perform the comparison between the reference data and the compared data.

        .. note::
            This function must return either of the following:

            * an iterable of differences between each data element (the iterable can be empty).
            * a mapping of differences between each data element in which the keys can be an
              element ID or a column name (the mapping can be empty).
            * a boolean indicating whether the files are different (`True`) or not (`False`).
        """

    def filter(self, differences, **kwargs):
        """Define a filter to remove specific elements from the result differences."""
        return differences

    def format_diff(self, difference, **kwargs):
        """Format one element difference."""
        return difference

    def sort(self, differences, **kwargs):
        """Sort the element differences."""
        return sorted(differences)

    def concatenate(self, differences, **kwargs):
        """Concatenate the differences."""
        return "\n".join(differences)

    def report(
        self,
        ref_file,
        comp_file,
        formatted_differences,
        diff_args,
        diff_kwargs,
        load_kwargs=None,
        format_data_kwargs=None,
        filter_kwargs=None,
        format_diff_kwargs=None,
        sort_kwargs=None,
        concat_kwargs=None,
        **kwargs,
    ):  # pylint: disable=too-many-arguments
        """Create a report from the formatted differences.

        .. note::
            This function must return a formatted report of the differences (usually as a string
            but it can be any type). If the passed differences are ``None``, ``False`` or an empty
            collection, the report should return ``False`` to state that the files are not
            different.
        """
        return diff_msg_formatter(
            ref_file,
            comp_file,
            formatted_differences,
            diff_args,
            diff_kwargs,
            load_kwargs=load_kwargs,
            format_data_kwargs=format_data_kwargs,
            filter_kwargs=filter_kwargs,
            format_diff_kwargs=format_diff_kwargs,
            sort_kwargs=sort_kwargs,
            concat_kwargs=concat_kwargs,
            report_kwargs=kwargs,
        )

    def __call__(
        self,
        ref_file,
        comp_file,
        *diff_args,
        return_raw_diffs=False,
        load_kwargs=None,
        format_data_kwargs=None,
        filter_kwargs=None,
        format_diff_kwargs=None,
        sort_kwargs=None,
        concat_kwargs=None,
        report_kwargs=None,
        **diff_kwargs,
    ):
        """Perform the comparison between the reference file and the compared file.

        .. note::
            The workflow is the following:

            * call :meth:`dir_content_diff.base_comparators.BaseComparator.load()` to load the
              reference file.
            * call :meth:`dir_content_diff.base_comparators.BaseComparator.load()` to load the
              compared file.
            * call :meth:`dir_content_diff.base_comparators.BaseComparator.format_data()` to format
              the data from the compared file.
            * call :meth:`dir_content_diff.base_comparators.BaseComparator.diff()` to compute the
              differences.
            * if ``return_raw_diffs``, the diffs are returned at this step.
            * if the diffs are not just a boolean, the collection is:
                * filtered by calling
                  :meth:`dir_content_diff.base_comparators.BaseComparator.filter()`.
                * formatted by calling
                  :meth:`dir_content_diff.base_comparators.BaseComparator.format_diff()` on each
                  element.
                * sorted by calling
                  :meth:`dir_content_diff.base_comparators.BaseComparator.sort()`.
                * concatenated into one string by calling
                  :meth:`dir_content_diff.base_comparators.BaseComparator.concatenate()`.
            * a report is generated by calling
              :meth:`dir_content_diff.base_comparators.BaseComparator.report()`.
        """
        if load_kwargs is None:
            load_kwargs = self._default_load_kwargs
        if format_data_kwargs is None:
            format_data_kwargs = self._default_format_data_kwargs
        if not diff_kwargs:
            diff_kwargs = self._default_diff_kwargs
        if filter_kwargs is None:
            filter_kwargs = self._default_filter_kwargs
        if format_diff_kwargs is None:
            format_diff_kwargs = self._default_format_diff_kwargs
        if sort_kwargs is None:
            sort_kwargs = self._default_sort_kwargs
        if concat_kwargs is None:
            concat_kwargs = self._default_concat_kwargs
        if report_kwargs is None:
            report_kwargs = self._default_report_kwargs

        # Reset current state
        self.current_state = {}

        # Load data
        ref = self.load(ref_file, **load_kwargs)
        comp = self.load(comp_file, **load_kwargs)

        # Format compared data
        formatted_comp = self.format_data(comp, ref=ref, **format_data_kwargs)

        # Compute the difference
        diffs = self.diff(ref, formatted_comp, *diff_args, **diff_kwargs)

        # Return raw differences if required
        if return_raw_diffs:
            return diffs

        # Format the difference elements
        if not diffs:
            formatted_diffs = False
        elif diffs is True:
            formatted_diffs = diffs
        else:
            filtered_diffs = self.filter(diffs, **filter_kwargs)
            if hasattr(filtered_diffs, "items"):
                formatted_diffs = self.concatenate(
                    self.sort(
                        [self.format_diff(i, **format_diff_kwargs) for i in filtered_diffs.items()],
                        **sort_kwargs,
                    ),
                    **concat_kwargs,
                )
            else:
                formatted_diffs = self.concatenate(
                    self.sort(
                        [self.format_diff(i, **format_diff_kwargs) for i in filtered_diffs],
                        **sort_kwargs,
                    ),
                    **concat_kwargs,
                )

        # Build the report
        return self.report(
            ref_file,
            comp_file,
            formatted_diffs,
            diff_args,
            diff_kwargs,
            load_kwargs=load_kwargs,
            format_data_kwargs=format_data_kwargs,
            filter_kwargs=filter_kwargs,
            format_diff_kwargs=format_diff_kwargs,
            sort_kwargs=sort_kwargs,
            concat_kwargs=concat_kwargs,
            **report_kwargs,
        )

    def __eq__(self, other):
        """Compare 2 :class:`dir_content_diff.base_comparators.BaseComparator` instances."""
        if type(self) is not type(other) or self.__dict__.keys() != other.__dict__.keys():
            return False

        for k, v in self.__dict__.items():
            if other.__dict__[k] != v:
                return False

        return True


class DefaultComparator(BaseComparator):
    """The comparator used by default when none is registered for a given extension.

    This comparator only performs a binary comparison of the files.
    """

    def diff(self, ref, comp, *args, **kwargs):
        """Compare binary data.

        This function calls :func:`filecmp.cmp`, read the doc of this function for details on
        args and kwargs.
        """
        return not filecmp.cmp(ref, comp)


class DictComparator(BaseComparator):
    """Comparator for dictionaries."""

    _ACTION_MAPPING = {
        "add": "Added the value(s) '{value}' in the '{key}' key.",
        "change": "Changed the value of '{key}' from {value[0]} to {value[1]}.",
        "remove": "Removed the value(s) '{value}' from '{key}' key.",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._format_mapping = {
            "add": self._format_add_value,
            "remove": self._format_remove_value,
            "change": self._format_change_value,
        }

    @staticmethod
    def _format_key(key):
        if isinstance(key, str):
            key = key.split(".")
        if key == [""]:
            key = []
        return "".join(f"[{k}]" for k in key)

    @staticmethod
    def _format_add_value(value):
        return json.dumps(dict(sorted(value)))

    @staticmethod
    def _format_remove_value(value):
        return json.dumps(dict(sorted(value)))

    @staticmethod
    def _format_change_value(value):
        value = list(value)
        for num, i in enumerate(value):
            if isinstance(i, str):
                value[num] = f"'{i}'"
            else:
                value[num] = str(i)
        return value

    def diff(self, ref, comp, *args, **kwargs):
        """Compare 2 dictionaries.

        This function calls :func:`dictdiffer.diff` to compare the dictionaries, read the doc of
        this function for details on args and kwargs.

        Keyword Args:
            tolerance (float): Relative threshold to consider when comparing two float numbers.
            absolute_tolerance (float): Absolute threshold to consider when comparing
                two float numbers.
            ignore (set[list]): Set of keys that should not be checked.
            path_limit (list[str]): List of path limit tuples or :class:`dictdiffer.utils.PathLimit`
                object to limit the diff recursion depth.
        """
        if len(args) > 5:
            dot_notation = args[5]
            args = args[:5] + args[6:]
        else:
            dot_notation = kwargs.pop("dot_notation", False)
        kwargs["dot_notation"] = dot_notation
        return list(dictdiffer.diff(ref, comp, *args, **kwargs))

    def format_diff(self, difference):
        """Format one element difference."""
        action, key, value = difference
        return self._ACTION_MAPPING[action].format(
            key=self._format_key(key),
            value=self._format_mapping[action](value),
        )


class JsonComparator(DictComparator):
    """Comparator for JSON files.

    This comparator is based on the :class:`DictComparator` and uses the same parameters.
    """

    def load(self, path):
        """Open a JSON file."""
        with open(path) as file:  # pylint: disable=unspecified-encoding
            data = json.load(file)
        return data


class YamlComparator(DictComparator):
    """Comparator for YAML files.

    This comparator is based on the :class:`DictComparator` and uses the same parameters.
    """

    def load(self, path):
        """Open a YAML file."""
        with open(path) as file:  # pylint: disable=unspecified-encoding
            data = yaml.full_load(file)
        return data


class XmlComparator(DictComparator):
    """Comparator for XML files.

    This comparator is based on the :class:`DictComparator` and uses the same parameters.

    .. warning:: The XML files must have only one root.

    .. note::

        If the type attributes are given in the XML file, the values will be automatically casted
        to Python types. For the lists, each item must be in an separated entry.
        Here is an example of such XML data:

        .. code-block:: xml

            <?xml version="1.0" encoding="UTF-8" ?>
            <root>
                <int_value type="int">1</int_value>
                <simple_list type="list">
                    <item type="int">1</item>
                    <item type="float">2.5</item>
                    <item type="str">str_val</item>
                </simple_list>
            </root>
    """

    def load(self, path):  # pylint: disable=arguments-differ
        """Open a XML file."""
        with open(path, encoding="utf-8") as file:
            data = self.xmltodict(file.read())
        return data

    @staticmethod
    def _cast_from_attribute(text, attr):
        """Converts XML text into a Python data format based on the tag attribute."""
        if "type" not in attr:
            return text
        value_type = attr.get("type", "").lower()
        if value_type == "str":
            res = str(text)
        elif value_type == "int":
            res = int(text)
        elif value_type == "float":
            res = float(text)
        elif value_type == "bool":
            if str(text).lower() == "true":
                res = True
            elif str(text).lower() == "false":
                res = False
            else:
                raise ValueError("Bool attributes expect 'true' or 'false'.")
        elif value_type == "list":
            res = []
        elif value_type == "dict":
            res = {}
        elif value_type == "null":
            res = None
        else:
            raise TypeError(
                "Unsupported type. "
                "Only 'str', 'int', 'float', 'bool', 'list', 'dict', and 'None' are supported."
            )
        return res

    @staticmethod
    def add_to_output(obj, child):
        """Add entry from :class:`xml.etree.ElementTree.Element` object into the given object."""
        if isinstance(obj, dict):
            obj.update({child.tag: XmlComparator._cast_from_attribute(child.text, child.attrib)})
            for sub in child:
                XmlComparator.add_to_output(obj[child.tag], sub)
        elif isinstance(obj, list):
            obj.append(XmlComparator._cast_from_attribute(child.text, child.attrib))
            for sub in child:
                XmlComparator.add_to_output(obj[-1], sub)

    @staticmethod
    def xmltodict(obj):
        """Converts an XML string into a Python object based on each tag's attribute."""
        root = ElementTree.fromstring(obj)
        output = {}

        for child in root:
            XmlComparator.add_to_output(output, child)
        return {root.tag: output}


class PdfComparator(BaseComparator):
    """Comparator for PDF files."""

    def diff(self, ref, comp, *args, **kwargs):
        """Compare data from two PDF files.

        This function calls the `diff_pdf_visually.pdfdiff() <https://github.com/bgeron/diff-pdf-
        visually/blob/21e85f1db1bdaee5c0e8e0b730771d6c4e8c3e44/diff_pdf_visually/diff.py#L83>`_
        function, read the doc of this function for details on args and kwargs.
        It compares the visual aspects of the PDF files, ignoring the invisible content (e.g. file
        header or invisible things like white font on white background). The PDF files are converted
        into images using ``ImageMagick`` and then these images are compared.

        Keyword Args:
            threshold (int): The threshold used to compare the images.
            tempdir (pathlib.Path): Empty directory where the temporary images will be exported.
            dpi (int): The resolution used to convert the PDF files into images.
            verbosity (int): The log verbosity.
            max_report_pagenos (int): Only this number of the different pages will be logged (only
                used if the verbosity is greater than 1).
            num_threads (int): If set to 2 (the default), the image conversion are processed in
                parallel. If set to 1 it is processed sequentially.
        """
        return not pdfdiff(ref, comp, *args, **kwargs)



"""Test the Pandas extension of the ``dir-content-diff`` package."""
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=unused-argument
# pylint: disable=use-implicit-booleaness-not-comparison
import re

import pandas as pd
import pytest

import dir_content_diff
import dir_content_diff.pandas
from dir_content_diff import assert_equal_trees
from dir_content_diff import compare_trees


class TestRegistry:
    """Test the internal registry."""

    def test_pandas_register(self, registry_reseter):
        """Test registering the pandas plugin."""
        assert dir_content_diff.get_comparators() == {
            None: dir_content_diff.DefaultComparator(),
            ".json": dir_content_diff.JsonComparator(),
            ".pdf": dir_content_diff.PdfComparator(),
            ".xml": dir_content_diff.XmlComparator(),
            ".yaml": dir_content_diff.YamlComparator(),
            ".yml": dir_content_diff.YamlComparator(),
        }

        dir_content_diff.pandas.register()
        assert dir_content_diff.get_comparators() == {
            None: dir_content_diff.DefaultComparator(),
            ".json": dir_content_diff.JsonComparator(),
            ".pdf": dir_content_diff.PdfComparator(),
            ".xml": dir_content_diff.XmlComparator(),
            ".yaml": dir_content_diff.YamlComparator(),
            ".yml": dir_content_diff.YamlComparator(),
            ".csv": dir_content_diff.pandas.CsvComparator(),
            ".tsv": dir_content_diff.pandas.CsvComparator(),
            ".h4": dir_content_diff.pandas.HdfComparator(),
            ".h5": dir_content_diff.pandas.HdfComparator(),
            ".hdf": dir_content_diff.pandas.HdfComparator(),
            ".hdf4": dir_content_diff.pandas.HdfComparator(),
            ".hdf5": dir_content_diff.pandas.HdfComparator(),
        }


@pytest.fixture
def pandas_registry_reseter(registry_reseter):
    """Register the pandas plugin and then reset the registry after the test."""
    dir_content_diff.pandas.register()


@pytest.fixture
def ref_hdf5(empty_ref_tree):
    """The reference HDF5 file."""
    ref_data = {
        "col_a": [1, 2, 3],
        "col_b": ["a", "b", "c"],
        "col_c": [4, 5, 6],
    }
    df = pd.DataFrame(ref_data, index=["idx1", "idx2", "idx3"])
    filename = empty_ref_tree / "file.h5"
    df.to_hdf(filename, key="data", index=True)
    return filename


@pytest.fixture
def res_hdf5_equal(ref_hdf5, empty_res_tree):
    """The result hdf5 file equal to the reference."""
    df = pd.read_hdf(ref_hdf5, index_col="index")
    filename = empty_res_tree / "file.h5"
    df.to_hdf(filename, key="data", index=True)
    return filename


@pytest.fixture
def res_hdf5_diff(ref_hdf5, empty_res_tree):
    """The result hdf5 file different from the reference."""
    df = pd.read_hdf(ref_hdf5, index_col="index")
    df.loc["idx1", "col_a"] *= 10
    df.loc["idx2", "col_b"] += "_new"
    filename = empty_res_tree / "file.h5"
    df.to_hdf(filename, key="data", index=True)
    return filename


class TestEqualTrees:
    """Tests that should return no difference."""

    def test_diff_tree(
        self, ref_tree, ref_csv, res_tree_equal, res_csv_equal, pandas_registry_reseter
    ):
        """Test that no difference is returned."""
        res = compare_trees(ref_tree, res_tree_equal)
        assert res == {}

    def test_specific_args(
        self, ref_tree, ref_csv, res_tree_diff, res_csv_diff, pandas_registry_reseter
    ):
        """Test specific args."""
        specific_args = {
            "file.csv": {
                "atol": 100000,
                "ignore_columns": ["col_b"],
            }
        }
        res = compare_trees(ref_tree, res_tree_diff, specific_args=specific_args)

        # The CSV file is considered as equal thanks to the given kwargs
        assert len(res) == 4
        assert re.match(".*/file.csv.*", str(res)) is None

    def test_replace_pattern(
        self, ref_tree, ref_csv, res_tree_equal, res_csv_equal, pandas_registry_reseter
    ):
        """Test the feature to replace a given pattern in files."""
        # Add a column with paths in the CSV file
        ref_df = pd.read_csv(ref_csv, index_col="index")
        ref_df["test_path"] = "relative_path/test.file"
        relative_path = "relative_path/test.file"
        absolute_path = str(res_tree_equal / relative_path)
        path_data = (
            f"Some text before the first path: {relative_path} some text after the first path.\n"
            f"Some text before the second path: {relative_path} some text after the second path."
        )
        ref_df["test_data_with_path"] = path_data
        ref_df["test_path_only_in_ref"] = relative_path
        ref_df.to_csv(ref_csv, index=True, index_label="index")

        res_df = pd.read_csv(res_csv_equal, index_col="index")
        res_df["test_path"] = absolute_path
        res_df["test_data_with_path"] = path_data.replace(relative_path, absolute_path, 1)
        res_df["test_path_only_in_res"] = absolute_path
        res_df.to_csv(res_csv_equal, index=True, index_label="index")

        # Check that the missing column is found
        specific_args = {
            "file.csv": {
                "format_data_kwargs": {
                    "replace_pattern": {
                        (str(res_tree_equal), ""): [
                            "test_path",
                            "test_data_with_path",
                            "test_path_only_in_ref",
                            "test_path_only_in_res",
                        ]
                    }
                }
            }
        }
        res = compare_trees(ref_tree, res_tree_equal, specific_args=specific_args)

        assert len(res) == 1
        res_csv = res["file.csv"]
        match_res = re.match(
            r"The files '\S*/ref/file.csv' and '\S*/res/file.csv' are different:\n"
            r"Kwargs used for formatting data: {'replace_pattern': {.*}}\n\n"
            r"Column 'test_path_only_in_ref': The column is missing in the compared DataFrame, "
            r"please fix the 'replace_pattern' argument.\n\n"
            r"Column 'test_path_only_in_res': The column is missing in the reference DataFrame, "
            r"please fix the 'replace_pattern' argument.",
            res_csv,
        )
        assert match_res is not None

        # Test with regex flag
        specific_args = {
            "file.csv": {
                "format_data_kwargs": {
                    "replace_pattern": {
                        (str(res_tree_equal), "", re.DOTALL): [
                            "test_path",
                            "test_data_with_path",
                            "test_path_only_in_ref",
                            "test_path_only_in_res",
                        ]
                    }
                }
            }
        }
        res = compare_trees(ref_tree, res_tree_equal, specific_args=specific_args)

        assert len(res) == 1
        res_csv = res["file.csv"]
        match_res = re.match(
            r"The files '\S*/ref/file.csv' and '\S*/res/file.csv' are different:\n"
            r"Kwargs used for formatting data: {'replace_pattern': {.*}}\n\n"
            r"Column 'test_path_only_in_ref': The column is missing in the compared DataFrame, "
            r"please fix the 'replace_pattern' argument.\n\n"
            r"Column 'test_path_only_in_res': The column is missing in the reference DataFrame, "
            r"please fix the 'replace_pattern' argument.",
            res_csv,
        )
        assert match_res is not None

        # Test with only nan values in column
        for df_path in [ref_csv, res_csv_equal]:
            df = pd.read_csv(df_path, index_col="index")
            df["test_path"] = None
            df.to_csv(df_path, index=True, index_label="index")

        res = compare_trees(ref_tree, res_tree_equal, specific_args=specific_args)

        assert len(res) == 1
        res_csv = res["file.csv"]
        match_res = re.match(
            r"The files '\S*/ref/file.csv' and '\S*/res/file.csv' are different:\n"
            r"Kwargs used for formatting data: {'replace_pattern': {.*}}\n\n"
            r"Column 'test_path_only_in_ref': The column is missing in the compared DataFrame, "
            r"please fix the 'replace_pattern' argument.\n\n"
            r"Column 'test_path_only_in_res': The column is missing in the reference DataFrame, "
            r"please fix the 'replace_pattern' argument.",
            res_csv,
        )
        assert match_res is not None

    def test_hdf5_comparator(
        self, empty_ref_tree, empty_res_tree, res_hdf5_equal, pandas_registry_reseter
    ):
        """Test the comparator for HDF5 files."""
        assert_equal_trees(empty_ref_tree, empty_res_tree, export_formatted_files=True)
        ref_files = list(empty_ref_tree.rglob("*"))
        res_files = list(empty_res_tree.rglob("*"))
        assert len(ref_files) == len(res_files)
        for i, j in zip(ref_files, res_files):
            assert pd.read_hdf(i).equals(pd.read_hdf(j))


class TestDiffTrees:
    """Tests that should return differences."""

    def test_diff_tree(
        self, ref_tree, ref_csv, res_tree_diff, res_csv_diff, csv_diff, pandas_registry_reseter
    ):
        """Test that the returned differences are correct."""
        res = compare_trees(ref_tree, res_tree_diff)

        assert len(res) == 5
        res_csv = res["file.csv"]
        match_res = re.match(csv_diff, res_csv)
        assert match_res is not None

    def test_read_csv_kwargs(
        self, ref_tree, ref_csv, res_tree_diff, res_csv_diff, csv_diff, pandas_registry_reseter
    ):
        """Test specific args for the CSV reader."""
        specific_args = {
            "file.csv": {"load_kwargs": {"header": None, "skiprows": 1, "prefix": "col_"}}
        }
        res = compare_trees(ref_tree, res_tree_diff, specific_args=specific_args)

        assert len(res) == 5
        res_csv = res["file.csv"]
        kwargs_msg = (
            "Kwargs used for loading data: {'header': None, 'skiprows': 1, 'prefix': 'col_'}\n"
        )
        assert kwargs_msg in res_csv
        match_res = re.match(
            csv_diff.replace("col_a", "col_1").replace("col_b", "col_2"),
            res_csv.replace(kwargs_msg, ""),
        )
        assert match_res is not None

    def test_missing_column(
        self, ref_tree, ref_csv, res_tree_diff, res_csv_equal, csv_diff, pandas_registry_reseter
    ):
        """Test the behavior with missing columns in CSV files."""
        # Rename a column from the CSV file
        df = pd.read_csv(res_csv_equal, index_col="index")
        df.rename(columns={"col_c": "new_col_c"}, inplace=True)
        df.to_csv(res_csv_equal, index=True, index_label="index")

        # Check that the missing column is found
        res = compare_trees(ref_tree, res_tree_diff)

        assert len(res) == 1
        res_csv = res["file.csv"]
        match_res = re.match(
            r"The files '\S*/ref/file.csv' and '\S*/res/file.csv' are different:\n\n"
            r"Column 'col_c': The column is missing in the compared DataFrame.\n\n"
            r"Column 'new_col_c': The column is missing in the reference DataFrame.",
            res_csv,
        )
        assert match_res is not None

    def test_hdf5_comparator(
        self, empty_ref_tree, empty_res_tree, res_hdf5_diff, pandas_registry_reseter
    ):
        """Test the comparator for HDF5 files."""
        res = compare_trees(empty_ref_tree, empty_res_tree)

        assert len(res) == 1
        res_hdf = res["file.h5"]
        match_res = re.match(
            r"The files '\S*/ref/file.h5' and '\S*/res/file.h5' are different:\n\n"
            r"Column 'col_a': Series are different\n\n"
            r"Series values are different \(33\.33333 %\)\n"
            r"\[index\]: \[idx1, idx2, idx3\]\n"
            r"\[left\]:  \[1, 2, 3\]\n\[right\]: \[10, 2, 3\]\n\n"
            r"Column 'col_b': Series are different\n\n"
            r"Series values are different \(33\.33333 %\)\n"
            r"\[index\]: \[idx1, idx2, idx3\]\n"
            r"\[left\]:  \[a, b, c\]\n\[right\]: \[a, b_new, c\]",
            res_hdf,
        )
        assert match_res is not None



"""Some utils used by the ``dir-content-diff`` package."""
import re

_ext_pattern = re.compile(r"\.?(.*)")


def format_ext(ext):
    """Ensure that the given extension string begins with a dot.

    Args:
        ext(str): The extension to format.

    Returns:
        The formatted extension.
    """
    match = re.match(_ext_pattern, ext)
    return f".{match.group(1)}"


def diff_msg_formatter(
    ref,
    comp,
    reason=None,
    diff_args=None,
    diff_kwargs=None,
    load_kwargs=None,
    format_data_kwargs=None,
    filter_kwargs=None,
    format_diff_kwargs=None,
    sort_kwargs=None,
    concat_kwargs=None,
    report_kwargs=None,
):  # pylint: disable=too-many-arguments
    """Format a difference message.

    Args:
        ref (str): The path to the reference file.
        comp (str): The path to the compared file.
        reason (bool or str): If the reason is False, False is returned. If it is a str, a formatted
            message is returned.
        diff_args (list): (optional) The args used for the comparison.
        diff_kwargs (list): (optional) The kwargs used for the comparison.
        load_kwargs (dict): The kwargs used for loading the data.
        format_data_kwargs (dict): The kwargs used for formatting the data.
        filter_kwargs (dict): The kwargs used for filtering the differences.
        format_diff_kwargs (dict): The kwargs used for formatting the differences.
        sort_kwargs (dict): The kwargs used for sorting the differences.
        concat_kwargs (dict): The kwargs used for concatenating the differences.
        report_kwargs (dict): The kwargs used for reporting the differences.

    Returns:
        False or the difference message.
    """
    if not reason:
        return False

    if reason is not None and reason is not True:
        reason_used = f"{reason}"
    else:
        reason_used = ""

    if diff_args:
        args_used = f"Args used for computing differences: {list(diff_args)}\n"
    else:
        args_used = ""

    def format_kwargs(kwargs, name):
        if kwargs:
            return f"Kwargs used for {name}: {kwargs}\n"
        return ""

    diff_kwargs_used = format_kwargs(diff_kwargs, "computing differences")
    load_kwargs_used = format_kwargs(load_kwargs, "loading data")
    format_data_kwargs_used = format_kwargs(format_data_kwargs, "formatting data")
    filter_kwargs_used = format_kwargs(filter_kwargs, "filtering differences")
    format_diff_kwargs_used = format_kwargs(format_diff_kwargs, "formatting differences")
    sort_kwargs_used = format_kwargs(sort_kwargs, "sorting differences")
    concat_kwargs_used = format_kwargs(concat_kwargs, "concatenating differences")
    report_kwargs_used = format_kwargs(report_kwargs, "reporting differences")

    kwargs_used = "\n".join(
        i
        for i in [
            load_kwargs_used,
            format_data_kwargs_used,
            diff_kwargs_used,
            filter_kwargs_used,
            format_diff_kwargs_used,
            sort_kwargs_used,
            concat_kwargs_used,
            report_kwargs_used,
        ]
        if i
    )

    eol = "."
    if reason_used or args_used or kwargs_used:
        eol = ":\n"

    return (
        f"The files '{ref}' and '{comp}' are different{eol}"
        f"{args_used}"
        f"{kwargs_used}"
        f"{reason_used}"
    )



"""Setup for the dir-content-diff package."""
from pathlib import Path

from setuptools import find_packages
from setuptools import setup

reqs = [
    "dictdiffer",
    "diff_pdf_visually>=1.5.1",
    "PyYaml",
]
doc_reqs = [
    "m2r2",
    "mistune<2",
    "sphinx",
    "sphinx-bluebrain-theme",
]
pandas_reqs = [
    "pandas",
    "tables",
]
test_reqs = [
    "coverage",
    "dicttoxml",
    "matplotlib",
    "rst2pdf>=0.99",
    "pytest",
    "pytest-html",
]

setup(
    name="dir-content-diff",
    author="Blue Brain Project, EPFL",
    description="Simple tool to compare directory contents.",
    long_description=Path("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    url="https://dir-content-diff.readthedocs.io",
    project_urls={
        "Tracker": "https://github.com/BlueBrain/dir-content-diff/issues",
        "Source": "https://github.com/BlueBrain/dir-content-diff",
    },
    license="Apache License 2.0",
    packages=find_packages(include=["dir_content_diff"]),
    python_requires=">=3.8",
    use_scm_version=True,
    setup_requires=[
        "setuptools_scm",
    ],
    install_requires=reqs,
    extras_require={
        "pandas": pandas_reqs,
        "docs": doc_reqs + pandas_reqs,
        "test": test_reqs + pandas_reqs,
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    include_package_data=True,
    entry_points={
        "pytest11": ["dir-content-diff = dir_content_diff.pytest_plugin"],
    },
)



"""Test the version of the ``dir-content-diff`` package."""
import pkg_resources

import dir_content_diff


def test_version():
    """Test the version of the dir-content-diff package."""
    pkg_version = pkg_resources.get_distribution("dir-content-diff").version
    assert dir_content_diff.__version__ == pkg_version
